// todo: move outside of this gadget
var BROWSER_URL = "http://localhost:9222";
var BROWSER_USERNAME = "ignored";
var BROWSER_PASSWORD = "ignored";

(function(window, rJS, RSVP) {
	"use strict";
	/**
	 * Makes an HTTP GET request to the browser to retrieve some JSON data.
	 *  
	 * @param {String} path The path of the json file.
	 * @returns The promise for the the JSON data as object. 
	 */
	function getBrowserJSON (path) {
		var callback = (resolve, reject) => {
			var request = new XMLHttpRequest();
			var url = (new URL(path, BROWSER_URL)).href;
			request.open("GET", url, true, BROWSER_USERNAME, BROWSER_PASSWORD);
			request.withCredentials = true;
			request.onload = () => {
				resolve(JSON.parse(request.responseText));
			}
			request.onerror = (error) => {
				reject("Could not connect to browser!");
			}
			request.send();
		}
		return RSVP.Queue().push(function () {
			return RSVP.Promise(callback);
		});
	}
	/**
	 * Returns the devtools API for the given domain. Domains are defined
	 * by the devtools protocol, e.g. Page. Each domain defines multiple
	 * methods, events, and types.
	 * 
	 * @param {object} domain The domain data from the devtools API.
	 * @returns The promise for the devtools API object and its name.
	 */
	 function getDevtoolsAPI (domain) {
		var callback = (resolve) => {
			var object = {};
			(domain.commands || []).forEach((method) => {
				addMethodAPI(object, domain, method);
			});
			(domain.events || []).forEach((event) => {
				addEventAPI(object, domain, event);
			});
			resolve([domain.domain, object]);
		}
		return RSVP.Queue().push(function () {
			return RSVP.Promise(callback);
		});
	}
	/**
	 * Adds the method API from the method JSON data to the object.
	 * 
	 * @param {object} object The container for the method.
	 * @param {object} domain The domain data from the protocol.
	 * @param {object} method The method data from the protocol.
	 */
	 function addMethodAPI (object, domain, method) {
		var result = (gadget, params) => {
			var callback = (resolve, reject) => {
				var sessionId = gadget.state.sessionId;
				var command = {
					sessionId,
					id: gadget.state.nextIndex,
					method: `${domain.domain}.${method.name}`,
					params: params,
				}
				console.log(command);
				var message = JSON.stringify(command);
				gadget.state.websocket.send(message);
				var handler = {
					sessionId,
					id: gadget.state.nextIndex,
					result: resolve,
					error: reject
				}
				gadget.state.handlers.add(handler);
				return gadget.changeState({update: true});
			}
			return RSVP.Queue().push(function () {
				return RSVP.Promise(callback);
			});
		}
		object[method.name] = result;
	}
	/**
	 * Adds the event API from the method JSON data to the object.
	 * 
	 * @param {object} object The container for the method.
	 * @param {object} domain The domain data from the protocol.
	 * @param {object} method The method data from the protocol.
	 */
	function addEventAPI (object, domain, method) {
		var result = (gadget, params) => {
			var callback = (resolve, reject) => {
				var sessionId = gadget.state.sessionId;
				var handler = {
					sessionId,
					method: `${domain.domain}.${method.name}`,
					result: resolve,
					error: reject
				}
				gadget.state.handlers.add(handler);
				return gadget.changeState({update: true});
			}
			return RSVP.Queue().push(function () {
				return RSVP.Promise(callback);
			});
		}
		object[method.name] = result;
	}
	/**
	 * Wraps the devtools API inside the gadget state which handels the
	 * connection with specific targets.
	 * 
	 * @param {*} gadget
	 * @param {*} devtoolsAPI 
	 */
	 function updateDevtools (gadget, devtoolsAPI) {
		var callback = () => {
			for (let [domain, methods] of Object.entries(devtoolsAPI)) {
				for (let [method, callback] of Object.entries(methods)) {
					devtoolsAPI[domain][method] = (params) => {
						return callback(gadget, params);
					};
				}
			}
			let attachToTarget = devtoolsAPI.Target.attachToTarget;
			devtoolsAPI.Target.attachToTarget = (params) => {
				var response;
				return attachToTarget(params)
					.push(function (result) {
						response = result;
						return gadget.changeState({sessionId: response.sessionId});
					})
					.push(function () {
						return response;
					})
			}
			let detachFromTarget = devtoolsAPI.Target.detachFromTarget;
			devtoolsAPI.Target.detachFromTarget = (params) => {
				return gadget.changeState({sessionId: undefined})
					.push(function () {
						return detachFromTarget(params);
					})
			}
			return gadget.changeState({devtools: devtoolsAPI});
		}
		return RSVP.Queue().push(callback);
	}
	/**
	 * Creates a new websocket that puts messages into the gadget's state.
	 * It also automatically reconnects on network errors.
	 * 
	 * @param {object} gadget The gadget that should receive messages. 
	 * @param {string} websocketUrl The url of the websocket.
	 */
	function updateWebsocket (gadget, websocketUrl) {
		var callback = (resolve) => {
			var websocket = new WebSocket(websocketUrl);
			websocket.onopen = resolve;
			websocket.onmessage = (message) => {
				gadget.state.messages.add(JSON.parse(message.data));
				gadget.changeState({nextIndex: gadget.state.nextIndex + 1});
			}
			websocket.onerror = (error) => {
				websocket.close();
				updateWebsocket(gadget, websocketUrl);
			}
			gadget.changeState({websocket: websocket});
		}
		return RSVP.Queue().push(function () {
			return RSVP.Promise(callback);
		});
	}
	/**
	 * Find the corresponding handler for the message and resolves it.
	 * 
	 * @param {object} message The message that contains the result data.
	 * @param {object} gadget The gadget that contains all the handlers.
	 */
	function handleMethodMessage(gadget, message) {
		for (let it = gadget.state.handlers.values(), handler = null; handler = it.next().value;) {
			if (handler.id != message.id) {
				continue;
			}
			
			gadget.state.handlers.delete(handler);
			gadget.state.messages.delete(message);

			if ('error' in message) {
				handler.error(message.error);
			} else {
				handler.result(message.result)
			}
		}
	}
	/**
	 * Find the corresponding handler for the event and resolves it.
	 * 
	 * @param {object} message The message that contains the event data.
	 * @param {object} gadget The gadget that contains all the handlers.
	 */
	function handleEventMessage(gadget, message) {
		for (let it = gadget.state.handlers.values(), handler = null; handler = it.next().value;) {
			setTimeout(function () {
				gadget.state.messages.delete(message);
			}, 5000);

			if (handler.method != message.method) {
				continue;
			}

			gadget.state.handlers.delete(handler);
			handler.result(message.params);
		}
	}

	rJS(window)
	.setState({
		update: false,
		websocket: null,
		sessionId: undefined,
		nextIndex: 0,
		handlers: null,
		messages: null,
		devtools: {},
	})
	.ready(function () {
		var gadget = this;
		gadget.state.handlers = new Set();
		gadget.state.messages = new Set();
		return getBrowserJSON("json/version")
			.push(function (result) {
				var browserUrl = result.webSocketDebuggerUrl;
				return updateWebsocket(gadget, browserUrl);
			})
			.push(function () {
				return getBrowserJSON("json/protocol")
			})
			.push(function (result) {
				var promises = result.domains.map(getDevtoolsAPI);
				return RSVP.all(promises);
			})
			.push(function (result) {
				var devtools = devtools = Object.fromEntries(result);
				return updateDevtools(gadget, devtools);
			})
	})
	.declareMethod("getDevtools", function () {
		var gadget = this;
		return RSVP.Queue().push(function () {
			return gadget.state.devtools;
		})
	})
	.onStateChange(function () {
		var gadget = this;
		gadget.state.update = false;

		for (let it = gadget.state.messages.values(), message = null; message = it.next().value;) {
			if (message.hasOwnProperty("id")) {
				handleMethodMessage(gadget, message);
			}
			if (message.hasOwnProperty("method")) {
				handleEventMessage(gadget, message);
			}
		}
	})
}(window, rJS, RSVP));